/*
    File Pear64.cpp

    Fancy diamond cut: Pear.

    Copyright (c) 2000 by OctoNus Ltd.
    All rights reserved.

    The file contains an implementation of two Pear cut modifications: with 64 and 32 points
    on the girdle.
    To compile this file as a part of DLL, define Pear64_EXPORTS preprocessor constant
    in the settings of Visual C++ project.
*/

#define DC_PARAMETERS_VERSION 2

#include "windows.h"
#include "math.h"
#include "DiamCut.h"

#define Pear64_API

#ifdef Pear64_EXPORTS

BOOL APIENTRY DllMain (HANDLE hModule, DWORD  ul_reason_for_call,
    LPVOID lpReserved )
{
    switch (ul_reason_for_call)
    {
        case DLL_PROCESS_ATTACH:
        case DLL_THREAD_ATTACH:
        case DLL_THREAD_DETACH:
        case DLL_PROCESS_DETACH:
             break;
    }
    return TRUE;
}

#endif // Pear64_EXPORTS

// Definition of data structure of cut parameters
// The structure does not depend on the number of points on the girdle,
// i.e. it is the same for Pear64 and Pear32
struct PearData
{
        int     version;        // data version
        // Main parameters
        double D;       // diameter
        double h;       // girdle thickness according to GIA (at the wide place)
        double alph;    // angle of main pavilion facets
        double beta;    // angle of main crown facets
        double crown;   // table size
        double pavilion;// height of lower pavilion facets
        // Girdle shape
        double rt;      // ratio of length to width
        double Lh;      // shift of the widest place
        double vLh;     // deviation of the shift of the widest place
        double vp;      // deviation of the angle on the point
        double vs;      // degree of deviation of the girdle from ellipse
        // Culet
        double culet;   // culet size
        double x, y;    // culet shift (y - towards the point of the Pear)
        // Crown facet picture
        double dSquare; // degree of deviation of crown facet picture from square
};

// Exported functions
Pear64_API int pearGetCuttingList (int * cuts, int n);
Pear64_API int pearGetCuttingInfo (void * data, int cut, int flavor);
Pear64_API int pearInitDiamond (DiamondCutting * diam, int cut);
Pear64_API int pearInitVertex (DiamondCutting * diam, int cut);
Pear64_API int pearIsCorrect (DiamondCutting * diam, int cut);
Pear64_API int pearGetParameter (double * value, int * param, int nparams,
        DiamondCutting * diam, int cut);
Pear64_API int pearSetParameter (double * value, int * param, int nparams,
        DiamondCutting * diam, int cut);


// Functions that are used internally
// List of parameters in the full diamond panel
static void FillPearFullParameter ( ParameterList * list );
// Recalculation of cut coordinates
static int PearInitVertex64 (DiamondCutting * diam);
static int PearInitVertex32 (DiamondCutting * diam);
// Getting / setting parameters
static int PearGetParameter ( double & value, int param, PearData * data );
static int PearSetParameter ( double value, int param, PearData * data );
// Girdle coordinate recalculation
static int FillGirdle_Pear( double u, double f, double s, double e, POINT2 * v );
static int MoveToPlanes ( const PHINDEX * S0, const PHINDEX * S1, POINT3 * vertex );
static double power ( double a, double b );
static int pear ( double u1, double d, double u2, double f, int n, POINT2 * v );


// Structure of Pear with 64 points on the girdle

static const PHINDEX IndexPear64 [ ] =
{
        // table
        0,  1,  2,  3,  4,  5,  6,  7,  0,
        // star facets
        0,  8,  1,  0,         1,  9,  2,  1,
        2, 10,  3,  2,         3, 11,  4,  3,
        4, 12,  5,  4,         5, 13,  6,  5,
        6, 14,  7,  6,         7, 15,  0,  7,
        // crown facets
        0, 15, 16,  8,  0,      1,  8, 24,  9,  1,
        2,  9, 32, 10,  2,      3, 10, 40, 11,  3,
        4, 11, 48, 12,  4,      5, 12, 56, 13,  5,
        6, 13, 64, 14,  6,      7, 14, 72, 15,  7,
        // upper crown facets (from 17 to 32)
         8, 16, 17, 18, 19, 20, 8,       8, 20, 21, 22, 23, 24, 8,
         9, 24, 25, 26, 27, 28, 9,       9, 28, 29, 30, 31, 32, 9,
        10, 32, 33, 34, 35, 36, 10,     10, 36, 37, 38, 39, 40, 10,
        11, 40, 41, 42, 43, 44, 11,     11, 44, 45, 46, 47, 48, 11,
        12, 48, 49, 50, 51, 52, 12,     12, 52, 53, 54, 55, 56, 12,
        13, 56, 57, 58, 59, 60, 13,     13, 60, 61, 62, 63, 64, 13,
        14, 64, 65, 66, 67, 68, 14,     14, 68, 69, 70, 71, 72, 14,
        15, 72, 73, 74, 75, 76, 15,     15, 76, 77, 78, 79, 16, 15,

        // girdle (from 33 to 96)

        16,  80,  81, 17, 16,
        17,  81,  82, 18, 17,
        18,  82,  83, 19, 18,
        19,  83,  84, 20, 19,
        20,  84,  85, 21, 20,
        21,  85,  86, 22, 21,
        22,  86,  87, 23, 22,
        23,  87,  88, 24, 23,
        24,  88,  89, 25, 24,
        25,  89,  90, 26, 25,
        26,  90,  91, 27, 26,
        27,  91,  92, 28, 27,
        28,  92,  93, 29, 28,
        29,  93,  94, 30, 29,
        30,  94,  95, 31, 30,
        31,  95,  96, 32, 31,
        32,  96,  97, 33, 32,
        33,  97,  98, 34, 33,
        34,  98,  99, 35, 34,
        35,  99, 100, 36, 35,
        36, 100, 101, 37, 36,
        37, 101, 102, 38, 37,
        38, 102, 103, 39, 38,
        39, 103, 104, 40, 39,
        40, 104, 105, 41, 40,
        41, 105, 106, 42, 41,
        42, 106, 107, 43, 42,
        43, 107, 108, 44, 43,
        44, 108, 109, 45, 44,
        45, 109, 110, 46, 45,
        46, 110, 111, 47, 46,
        47, 111, 112, 48, 47,
        48, 112, 113, 49, 48,
        49, 113, 114, 50, 49,
        50, 114, 115, 51, 50,
        51, 115, 116, 52, 51,
        52, 116, 117, 53, 52,
        53, 117, 118, 54, 53,
        54, 118, 119, 55, 54,
        55, 119, 120, 56, 55,
        56, 120, 121, 57, 56,
        57, 121, 122, 58, 57,
        58, 122, 123, 59, 58,
        59, 123, 124, 60, 59,
        60, 124, 125, 61, 60,
        61, 125, 126, 62, 61,
        62, 126, 127, 63, 62,
        63, 127, 128, 64, 63,
        64, 128, 129, 65, 64,
        65, 129, 130, 66, 65,
        66, 130, 131, 67, 66,
        67, 131, 132, 68, 67,
        68, 132, 133, 69, 68,
        69, 133, 134, 70, 69,
        70, 134, 135, 71, 70,
        71, 135, 136, 72, 71,
        72, 136, 137, 73, 72,
        73, 137, 138, 74, 73,
        74, 138, 139, 75, 74,
        75, 139, 140, 76, 75,
        76, 140, 141, 77, 76,
        77, 141, 142, 78, 77,
        78, 142, 143, 79, 78,
        79, 143,  80, 16, 79,
        // lower pavilion facets (from 97 to 112)
        144,  84,  83,  82,  81,  80, 144,
        144,  88,  87,  86,  85,  84, 144,
        145,  92,  91,  90,  89,  88, 145,
        145,  96,  95,  94,  93,  92, 145,
        146, 100,  99,  98,  97,  96, 146,
        146, 104, 103, 102, 101, 100, 146,
        147, 108, 107, 106, 105, 104, 147,
        147, 112, 111, 110, 109, 108, 147,
        148, 116, 115, 114, 113, 112, 148,
        148, 120, 119, 118, 117, 116, 148,
        149, 124, 123, 122, 121, 120, 149,
        149, 128, 127, 126, 125, 124, 149,
        150, 132, 131, 130, 129, 128, 150,
        150, 136, 135, 134, 133, 132, 150,
        144, 140, 139, 138, 137, 136, 144,
        144,  80, 143, 142, 141, 140, 144,
        // pavilion facets
         88, 144, 151, 152, 145,  88,
         96, 145, 152, 153, 146,  96,
        104, 146, 153, 154, 147, 104,
        112, 147, 154, 155, 148, 112,
        120, 148, 155, 156, 149, 120,
        128, 149, 156, 157, 150, 128,
        136, 150, 157, 151, 144, 136,
        // culet
        157, 156, 155, 154, 153, 152, 151, 157,
        // ending flag
        MAXPHINDEX
};

// Structure of Pear with 32 points on the girdle
static const PHINDEX IndexPear32 [ ] =
{
        // table
        0,  1,  2,  3,  4,  5,  6,  7,  0,
        // star facets
        0,  8,  1,  0,         1,  9,  2,  1,
        2, 10,  3,  2,         3, 11,  4,  3,
        4, 12,  5,  4,         5, 13,  6,  5,
        6, 14,  7,  6,         7, 15,  0,  7,
        // crown facets
        0, 15, 16,  8,  0,      1,  8, 20,  9,  1,
        2,  9, 24, 10,  2,      3, 10, 28, 11,  3,
        4, 11, 32, 12,  4,      5, 12, 36, 13,  5,
        6, 13, 40, 14,  6,      7, 14, 44, 15,  7,
        // upper crown facets (from 17 to 32)
         8, 16, 17, 18, 8,       8, 18, 19, 20, 8,
         9, 20, 21, 22, 9,       9, 22, 23, 24, 9,
        10, 24, 25, 26, 10,     10, 26, 27, 28, 10,
        11, 28, 29, 30, 11,     11, 30, 31, 32, 11,
        12, 32, 33, 34, 12,     12, 34, 35, 36, 12,
        13, 36, 37, 38, 13,     13, 38, 39, 40, 13,
        14, 40, 41, 42, 14,     14, 42, 43, 44, 14,
        15, 44, 45, 46, 15,     15, 46, 47, 16, 15,

        // girdle (from 33 to 64)

        16,  48,  49, 17, 16,
        17,  49,  50, 18, 17,
        18,  50,  51, 19, 18,
        19,  51,  52, 20, 19,
        20,  52,  53, 21, 20,
        21,  53,  54, 22, 21,
        22,  54,  55, 23, 22,
        23,  55,  56, 24, 23,
        24,  56,  57, 25, 24,
        25,  57,  58, 26, 25,
        26,  58,  59, 27, 26,
        27,  59,  60, 28, 27,
        28,  60,  61, 29, 28,
        29,  61,  62, 30, 29,
        30,  62,  63, 31, 30,
        31,  63,  64, 32, 31,
        32,  64,  65, 33, 32,
        33,  65,  66, 34, 33,
        34,  66,  67, 35, 34,
        35,  67,  68, 36, 35,
        36,  68,  69, 37, 36,
        37,  69,  70, 38, 37,
        38,  70,  71, 39, 38,
        39,  71,  72, 40, 39,
        40,  72,  73, 41, 40,
        41,  73,  74, 42, 41,
        42,  74,  75, 43, 42,
        43,  75,  76, 44, 43,
        44,  76,  77, 45, 44,
        45,  77,  78, 46, 45,
        46,  78,  79, 47, 46,
        47,  79,  48, 16, 47,

        // lower pavilion facets (from 65 to 80) (beginning with the point)
        80,  50,  49,  48,  80,
        80,  52,  51,  50,  80,
        81,  54,  53,  52,  81,
        81,  56,  55,  54,  81,
        82,  58,  57,  56,  82,
        82,  60,  59,  58,  82,
        83,  62,  61,  60,  83,
        83,  64,  63,  62,  83,
        84,  66,  65,  64,  84,
        84,  68,  67,  66,  84,
        85,  70,  69,  68,  85,
        85,  72,  71,  70,  85,
        86,  74,  73,  72,  86,
        86,  76,  75,  74,  86,
        80,  78,  77,  76,  80,
        80,  48,  79,  78,  80,

        // pavilion facets (from 81 to 87)
        52, 80, 87, 88, 81, 52,
        56, 81, 88, 89, 82, 56,
        60, 82, 89, 90, 83, 60,
        64, 83, 90, 91, 84, 64,
        68, 84, 91, 92, 85, 68,
        72, 85, 92, 93, 86, 72,
        76, 86, 93, 87, 80, 76,
        // culet (88)
        93, 92, 91, 90, 89, 88, 87, 93,
        // ending flag
        MAXPHINDEX
};


// Internal numeration of cuts in this module (library)
enum
{
        Pear64 = 1,
        Pear32
};

// Some mathematical defines

#define DEGREE          0.01745329251994
#define M_PI            3.14159265358979323846
#define M_PI_2          1.57079632679489661923
#define M_SQRT2         1.41421356237309504880
#define SQRT2           1.41421356237309504880


// Set of default parameters
static  PearData PearPars =
{
        1,              // version
        6,              // diameter
        0.0375,         // h - girdle thickness according to GIA (at the wide place)
        41*DEGREE,      // alph - angle of main pavilion facets
        35*DEGREE,      // beta - angle of main crown facets
        0.60,           // crown - table size
        0.18,           // pavilion - height of lower pavilion facets
        // Girdle shape
        1.5,            // rt - ratio of length to width
        0.34,           // Lh - shift of the widest place
        0,              // vLh - deviation of the shift of the widest place
        0,              // vp - deviation of the angle on the point
        0,              // vs - degree of deviation of the girdle from ellipse
        // Culet
        1e-5,           // culet - culet size
        0,0,            // x, y - culet shift (y - towards the point of the Pear)
        // Crown facet picture
        0               // dSquare - degree of deviation of crown facet picture from square

};


// Implementation of the interface of exported cuts

// Unique identifiers
// Attention! Creating cuts on the base of this example, ALWAYS redefine
// unique cut identifiers. Existance of two cuts with the same identifier
// will lead to collisions during software execution.
// {FE6BC6E2-1C29-11d4-8C66-00105A711E33}
static const CLSID CLSID_Pear64 = 
{ 0xfe6bc6e2, 0x1c29, 0x11d4, { 0x8c, 0x66, 0x0, 0x10, 0x5a, 0x71, 0x1e, 0x33 } };

// {FE6BC6E3-1C29-11d4-8C66-00105A711E33}
static const CLSID CLSID_Pear32 = 
{ 0xfe6bc6e3, 0x1c29, 0x11d4, { 0x8c, 0x66, 0x0, 0x10, 0x5a, 0x71, 0x1e, 0x33 } };


// Exported function: List of implemented cuts.
//      The function fills the 'cuts' list with cuts identifiers
// that are unique for this module.
// Returned value: number of cuts
// External name: GetCuttingList
// External name is different from the internal one to allow
//   direct linking of this module sometimes, without use of
//   LoadLibrary.Windows API
Pear64_API int pearGetCuttingList (int * cuts, int n)
{
    int i = 0;
    if (i < n) cuts [i++] = Pear64;
    if (i < n) cuts [i++] = Pear32;
    return i;
}

#define Parameter_vp     DC_UserParameterAngle
#define Parameter_Lh     DC_UserParameterPercent
#define Parameter_vLh    (DC_UserParameterPercent+1)
#define Parameter_vs     (DC_UserParameterPercent+2)
#define Parameter_dSquare (DC_UserParameterPercent+3)

// Exported function: getting misc. information on the cut
// External name: GetCuttingInfo
// Returned value: 1, if the requested information is returned
Pear64_API int pearGetCuttingInfo (void * data, int cut, int flavor)
{
    switch(flavor)
    {
        default: return 0;
        case DC_ClassDiamondCutting: //  
        {
            ClassDiamondCutting * info = (ClassDiamondCutting*) data;
            switch (cut)
            {
                default: return 0;
                case Pear64 :
                    info -> dataSize = sizeof(PearData);
                    info -> uid = CLSID_Pear64;
                    info -> version = 1;
                    info -> sides = IndexPear64;
                    info -> name = "Pear";
                    info -> shortname = "Pear";
                    break;
                case Pear32 :
                    info -> dataSize = sizeof(PearData);
                    info -> uid = CLSID_Pear32;
                    info -> version = 1;
                    info -> sides = IndexPear32;
                    info -> name = "Pear";
                    info -> shortname = "Pear32";
                    break;
            }
            break;
        }
        case DC_ParameterList: // parameter lists
        {
            ParameterList * list = (ParameterList *) data;
            if (cut != Pear64 && cut != Pear32) return 0;
            switch (list -> listType)
            {
                default: return 0;
                case DC_FullParameterList: // full parameter list
                    FillPearFullParameter ( list );
                    break;
            }
            break;
        }
        case DC_ParameterInfo: // information on the specific parameter
        {
            ParameterInfo * info = (ParameterInfo *) data;
            if (cut != Pear64 && cut != Pear32) return 0;
            switch (info->parameter)
            {
                case Parameter_vp:      info -> name = "vp"; break;
                case Parameter_Lh:      info -> name = "Lh"; break;
                case Parameter_vLh:     info -> name = "vL"; break;
                case Parameter_vs:      info -> name = "vs"; break;
                case Parameter_dSquare: info -> name = "Sq"; break;
            }
            break;
        }
    }
    return 1;
}

// Exported initialization of cut data
// External name: InitDiamond.
// Returned value: 1, if the initialization completed successfully
Pear64_API int pearInitDiamond (DiamondCutting * diam, int cut)
{
    switch (cut)
    {
        default: return 0;
        case Pear64 :
        case Pear32 :
        {
            // Default initialization simply copies parameter values
            // from PearPars structure
            PearData * data = (PearData*) diam -> data;
            *data = PearPars;
            break;
        }
    }
    return 1;
}


// Exported function of recalculating points' coordinates
// External name: InitVertex.
// Returned value: 1, if points are recalculated
Pear64_API int pearInitVertex (DiamondCutting * diam, int cut)
{
    switch (cut)
    {
        default: return 0;
        case Pear64: return PearInitVertex64(diam);
        case Pear32: return PearInitVertex32(diam);
    }
    return 1;
}

// Exported function of correctness of points' recalculation
// External name: IsCorrect.
// The returned value: 1, if the parameter set is correct, i.e. this
// parameters can be used for point recalculation and the diamond will be usable
Pear64_API int pearIsCorrect (DiamondCutting * diam, int cut)
{
    switch (cut)
    {
        default: return 0;
        case Pear64:
        case Pear32:
        {
            PearData & data = *(PearData*) diam -> data;
            // Simple check of parameters
            if ( ! ( data.D > 0 &&
                data.version <= 1 &&
                data.alph > 0 && data.alph < M_PI_2 &&
                data.beta > 0 && data.beta < M_PI_2 &&
                data.crown > 0 && data.crown < 1 &&
                data.pavilion < 1-1e-6 &&
                5e-6 < data.h &&
                1e-6 < data.rt &&
                data.culet > 1e-6 ) ) return 0;
            double m = (1 + SQRT2) / 2 * data.crown;
            if ( data.dSquare <= 0 ) m = m + data.dSquare * (m-1+data.crown);
            else m = m + data.dSquare * (1-m);
            if ( m <= data.crown ) return 0;
            int i;
            POINT2 girdle [64];
            if (FillGirdle_Pear ( data.rt, data.vp, data.rt * data.Lh - 0.5,
                data.vs, girdle ) != 0 ) return 0;
            // For simplicity, check when R = 1
            // Check if star points went out of girdle and crown
            LINE2 line [8]; // set of tangents to girdle points
            for ( i = 0; i < 8; i++ )
            {
                VECTOR2 dir = girdle [1+i*8] - girdle [(63+i*8)%64];
                POINT2 & p = girdle [i*8];
                line [i].norm = VECTOR2 (-dir.y, dir.x).fix ();
                line [i].dist = -(line [i].norm |p);
            }
            POINT2 g2 [8]; // points of intersection of previous tangents
            for ( i = 0; i < 8; i++ )
                g2 [i] = lines_intersect (line [i], line [(i+1)%8]);
            double offset = -2 * (data.vLh + data.rt * data.Lh - 0.5); // shift of the central point
            POINT2 upPoint (0, offset);
            POINT2 downPoint (-2*data.x, -2*data.y+offset);
            // Check if star points or imaginery culet goes out of the simplified girdle (8 point)
            for ( i = 0; i < 8; i++ )
            {
                POINT2 & p0 = girdle [i*8];
                POINT2 & p1 = girdle [(i*8+8)%64];
                LINE2 l (p0, p1);
                POINT2 star = upPoint + m * (g2 [i] - upPoint);
                if ((l.norm |star)+l.dist < 0 || (l.norm |downPoint)+l.dist < 0) return 0;
            }
            // check of pavilion height remains
            double pavilion = (girdle [0] - downPoint).abs ();
            if (pavilion > 1e-6) pavilion = data.pavilion / pavilion * (g2 [0]-downPoint).abs();
            else pavilion = data.pavilion;
            if (pavilion > 0.99) return 0;
            for (i = 0; i < 7; i++)
            {
                int i0 = (i > 3) ? i + 1 : i;
                int i1 = ((i > 2) ? i + 1 : i) + 1;
                POINT2 g = lines_intersect (line [i0], line [i1%8]);
                double pav = (g - downPoint).abs ();
                // Smart dependence (i0+i1)*4 gives the required number
                // of the point on the girdle
                int index = (i0+i1)*4;
                double len = (girdle [index] - downPoint).abs ();
                if (pav > 1e-6) pav = pavilion * len / pav;
                if ( pav < data.culet ) return 0;
            }
            // All tests performed. The parameter set is correct.
            break;
        }
    }
    return 1;
}

// Exported function of getting diamond parameter
// External name: GetParameter.
// Returned value: 1, if the parameter set is received successfully
Pear64_API int pearGetParameter (double * value, int * param, int nparams,
        DiamondCutting * diam, int cut)
{
    switch (cut)
    {
        default: return 0;
        case Pear64 :
        case Pear32 :
        {
            PearData * data = (PearData*) diam -> data;
            for (int i = 0; i < nparams; i++)
                if (!PearGetParameter(value [i], param [i], data)) return 0;
            break;
        }
    }
    return 1;
}

// Exported function of setting diamond parameter
// External name: SetParameter.
// Returned value: 1, if the parameter set is applied
// Unfortunately, setting a set of parameters is implemented by
// setting parameters one-by-one!
Pear64_API int pearSetParameter (double * value, int * param, int nparams,
        DiamondCutting * diam, int cut)
{
    switch (cut)
    {
        default: return 0;
        case Pear64 :
        case Pear32 :
        {
            PearData * data = (PearData*) diam -> data;
            for (int i = 0; i < nparams; i++)
                if (!PearSetParameter(value [i], param [i], data)) return 0;
            break;
        }
    }
    return 1;
}

// Recalculation of girdle points
static int FillGirdle_Pear (double u, double f, double s, double e, POINT2 * v)
{
    double u1 = 1. + s + s;
    if ( u1 < 0 ) return 3;
    double u2 = u + u - u1;
    if ( u2 < 0 ) return 1;
    double a = asin ( ( u2 + u2 ) / ( u2 * u2 + 1.) );
    int r = pear ( u1, e+e, u2, f+a, 64, v );
    if ( !r && v )
    {
        double d = 1. - v[0].y;
        for ( int i = 0; i < 64; ++ i ) v[i].y += d;
    }
    return r;
}

static double power ( double a, double b )
{
    if ( b == 0 ) return 1.;
    if ( a == 0 ) return 0.;
    double t = log ( a ) * b;
    if ( t < log ( 1e-300 ) ) return 0.;
    return exp ( t );
}

static int pear ( double u1, double d, double u2, double f, int n, POINT2 * v )
{
    if ( f >= M_PI/2 ) return 2;
    double t = sqrt ( 0.5 ) + d / sqrt ( 1.+ u1 * u1 );
    if ( t > 0.99 || t < 0.5 ) return 3;
    double p = log ( 0.5 ) / log ( t );
    t = u2 * tan ( f );
    if ( t <= 2.) return 2;
    if ( ! v ) return 0;
    int n2 = n / 2;
    int n4 = n / 4;
    double h = M_PI / n2;
	int i;
    for ( i = 0; i < n4; ++ i )
    {
        double x = sin ( i * h );
        double y = cos ( i * h );
        double w = power ( y, p ) + power ( x, p );
        w = 1. / power ( w, 1./p );
        v[i].x = w * x;
        v[i].y = w * y * u1;
    }
    double a = ( t - 1. ) / ( t - 2. );
    double x0 = a - 1.;
    double b = u2 * ( 1.- t ) / sqrt ( t * t - ( t + t ) );
    h = acos ( x0 / a ) / n4;
    for ( i = 0; i < n4; ++ i )
    {
        v[i+n4].x = a * cos ( i * h ) - x0;
        v[i+n4].y = b * sin ( i * h );
    }
    v[n2].x = 0.;
    v[n2].y = - u2;
    for ( i = 1; i < n2; ++ i )
    {
        v[i+n2].x = - v[n2-i].x;
        v[i+n2].y =   v[n2-i].y;
    }
    return 0;
}

// Internal function of recalculation Pear32 coordinates
static int PearInitVertex32 (DiamondCutting * diam)
{
    int i;
    POINT3 * vertex = diam->vertex;
    PearData & data = *(PearData*) diam -> data;
    double D = data.D;
    double R = D / 2;
    double r_tan_alph = R * tan ( data.alph ),
           r_tan_beta = R * tan ( data.beta );
    double H1 = -data.h * R;        // level of upper part of the girdle
    double H2 = data.h * R;         // level of lower part of the girdle
    POINT3 * girdle = & vertex [16];
    POINT2 girdle2 [64];
    FillGirdle_Pear ( data.rt, data.vp, data.rt * data.Lh - 0.5, data.vs, girdle2 );
    // Unfortunately, the previous function does not depend on D
    for (i = 0; i < 32; i++)
    {
        int off = (i*2+32)%64;
        girdle [i].x = -girdle2 [off].x * R;
        girdle [i].y = -girdle2 [off].y * R;
        girdle [i].z = H1;
    }
    // Fill the second level of the girdle
    for (i = 0; i < 32; i++)
    {
        girdle [i+32].x = girdle [i].x;
        girdle [i+32].y = girdle [i].y;
        girdle [i+32].z = H2;
    }
    // crown points are proportional girdle points relative to upPoint (it follows
    // from the assumption that all facets intersect in the one point)
    double offset = (data.vLh + data.rt * data.Lh - 0.5) * D; // shift of the central point
    POINT3 upPoint (0, offset, H1 - r_tan_beta);
    for ( i = 0; i < 8; i++ )
    {
        VECTOR3 dir = girdle [i*4] - upPoint;
        vertex [i] = upPoint + data.crown * dir;
    }
    // Points of intersection of main crown facets between them on the girdle level
    LINE2 line [8]; // tangents to main 8 facets on the girdle
    for ( i = 0; i < 8; i++ )
    {
        VECTOR3 dir = girdle [1+i*4] - girdle [(31+i*4)%32];
        POINT3 & p = girdle [i*4];
        line [i].norm = VECTOR2 (-dir.y, dir.x).fix ();
        line [i].dist = -(line [i].norm |POINT2(p.x, p.y));
    }
    POINT2 g2 [8]; // points of intersection of the previous tangents between them
    for ( i = 0; i < 8; i++ )
        g2 [i] = lines_intersect (line [i], line [(i+1)%8]);
    // star points are proportional to these points relative to upPoint
    double m = (1 + SQRT2) / 2 * data.crown;
    if ( data.dSquare <= 0 ) m = m + data.dSquare * (m-1+data.crown);
    else m = m + data.dSquare * (1-m);
    for ( i = 0; i < 8; i++ )
    {
        VECTOR3 dir = POINT3 (g2 [i].x, g2[i].y, H1) - upPoint;
        vertex [i+8] = upPoint + m * dir;
    }
    // culet points are proportional to girdle points relative to
    // shift point. But there are only 7 of them!!!
    // pavilion points are on different heights, so that they divide bones in equal
    // proportion when viewed from below
    POINT2 g7 [7];
    for (i = 1; i <= 6; i++) g7 [i] = g2 [i];
    g7 [0] = lines_intersect (line [7], line [1]);
    POINT3 downPoint (data.x*D, data.y*D + offset, H2 + r_tan_alph);
    POINT2 downPoint2 (downPoint.x, downPoint.y);
    // for the point 4 the correct ratio should be fulfilled
    double pavilion = (POINT2 (girdle [50].x, girdle [50].y) - downPoint2).abs ();
    if (pavilion > 1e-6) pavilion = data.pavilion / pavilion * (g7 [4]-downPoint2).abs();
    else pavilion = data.pavilion;
    for (i = 0; i < 7; i++)
    {
        POINT3 dir = POINT3 (g7 [i].x, g7 [i].y, H2) - downPoint;
        vertex [i+87] = downPoint + data.culet * dir;
        int index = (i == 0) ? 0 : 2;
        index += 32 + i * 4;
        double pav = (g7 [i] - downPoint2).abs ();
        if (pav > 1e-6)
        {
            pav = pavilion * (POINT2 (girdle [index].x, girdle [index].y) -
                downPoint2).abs () / pav;
        }
        vertex [i+80] = downPoint + pav * dir;
    }
    PHINDEX v0 [4], v1 [4];
    // and, finally, more precise determination of girdle points
    for ( i = 0; i < 16; i++ )
    {
        v0 [0] = 8 + i / 2;
        v0 [1] = 16 + i * 2;
        v0 [3] = 16 + (i * 2 + 2)%32;
        v1 [0] = v0 [0] + 72;
        v1 [1] = v0 [3] + 32;
        v1 [3] = v0 [1] + 32;
        v0 [2] = v0 [1] + 1;
        v1 [2] = v1 [3] + 1;
        MoveToPlanes (v1, v0, vertex);
    }
    return 0;
}

// Internal function of recalculation of Pear64 coordinates
static int PearInitVertex64 (DiamondCutting * diam)
{
    int i, j;
    POINT3 * vertex = diam->vertex;
    PearData & data = *(PearData*) diam -> data;
    double D = data.D;
    double R = D / 2;
    double r_tan_alph = R * tan ( data.alph ),
           r_tan_beta = R * tan ( data.beta );
    double H1 = -data.h * R;        // level of upper part of the girdle
    double H2 = data.h * R;         // level of lower part of the girdle
    POINT3 * girdle = & vertex [16];
    POINT2 girdle2 [64];
    FillGirdle_Pear ( data.rt, data.vp, data.rt * data.Lh - 0.5, data.vs, girdle2 );
    // Unfortunately, the previous function does not depend on D
    for (i = 0; i < 64; i++)
    {
        int off = (i+32)%64;
        girdle [i].x = -girdle2 [off].x * R;
        girdle [i].y = -girdle2 [off].y * R;
        girdle [i].z = H1;
    }
    // Fill the second level of the girdle
    for (i = 0; i < 64; i++)
    {
        girdle [i+64].x = girdle [i].x;
        girdle [i+64].y = girdle [i].y;
        girdle [i+64].z = H2;
    }
    // crown points are proportional to girdle points relative to upPoint (if follows
    // from the assumption that all facets intersect in one point)
    double offset = (data.vLh + data.rt * data.Lh - 0.5) * D; // shift of the central point
    POINT3 upPoint (0, offset, H1 - r_tan_beta);
    for ( i = 0; i < 8; i++ )
    {
        VECTOR3 dir = girdle [i*8] - upPoint;
        vertex [i] = upPoint + data.crown * dir;
    }
    // Points of intersection of main crown facets between them on the girdle level
    LINE2 line [8]; // tangents to main 8 points on the girdle
    for ( i = 0; i < 8; i++ )
    {
        VECTOR3 dir = girdle [1+i*8] - girdle [(63+i*8)%64];
        POINT3 & p = girdle [i*8];
        line [i].norm = VECTOR2 (-dir.y, dir.x).fix ();
        line [i].dist = -(line [i].norm |POINT2(p.x, p.y));
    }
    POINT2 g2 [8]; // points of intersection of the previous tangents between them
    for ( i = 0; i < 8; i++ )
        g2 [i] = lines_intersect (line [i], line [(i+1)%8]);
    // star points are proportional to these points relative to upPoint
    double m = (1 + SQRT2) / 2 * data.crown;
    if ( data.dSquare <= 0 ) m = m + data.dSquare * (m-1+data.crown);
    else m = m + data.dSquare * (1-m);
    for ( i = 0; i < 8; i++ )
    {
        VECTOR3 dir = POINT3 (g2 [i].x, g2[i].y, H1) - upPoint;
        vertex [i+8] = upPoint + m * dir;
    }
    // culet points are proportional to girdle poins relative to
    // the shift point. But there are only 7 of them!!!
    // pavilion points are at different height, so that they devide bones
    // when viewed from below
    POINT2 g7 [7];
    for (i = 1; i <= 6; i++) g7 [i] = g2 [i];
    g7 [0] = lines_intersect (line [7], line [1]);
    POINT3 downPoint (data.x*D, data.y*D + offset, H2 + r_tan_alph);
    POINT2 downPoint2 (downPoint.x, downPoint.y);
    // for the point 4 the correct ratio should be fulfilled
    double pavilion = (POINT2 (girdle [100].x, girdle [100].y) - downPoint2).abs ();
    if (pavilion > 1e-6) pavilion = data.pavilion / pavilion * (g7 [4]-downPoint2).abs();
    else pavilion = data.pavilion;
    for (i = 0; i < 7; i++)
    {
        POINT3 dir = POINT3 (g7 [i].x, g7 [i].y, H2) - downPoint;
        vertex [i+151] = downPoint + data.culet * dir;
        int index = (i == 0) ? 0 : 4;
        index += 64 + i * 8;
        double pav = (g7 [i] - downPoint2).abs ();
        if (pav > 1e-6)
        {
            pav = pavilion * (POINT2 (girdle [index].x, girdle [index].y) -
                downPoint2).abs () / pav;
        }
        vertex [i+144] = downPoint + pav * dir;
    }
    PHINDEX v0 [4], v1 [4];
    // and, finally, more precise determination of girdle points
    for ( i = 0; i < 16; i++ )
    {
        v0 [0] = 8 + i / 2;
        v0 [1] = 16 + i * 4;
        v0 [3] = 16 + (i * 4 + 4)%64;
        v1 [0] = v0 [0] + 136;
        v1 [1] = v0 [3] + 64;
        v1 [3] = v0 [1] + 64;
        for ( j = 1; j <= 3; j++ )
        {
            v0 [2] = v0 [1] + j;
            v1 [2] = v1 [3] - j + 4;
            MoveToPlanes (v1, v0, vertex);
        }
    }

    return 0;
}

// The list of parameters in the full cut panel (DC_FullParameterList)
static void FillPearFullParameter ( ParameterList * list )
{
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_CrownAngle;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_PavilionAngle;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_Diameter;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_CrownHeight;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_GirdleThickness;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_PavilionDepth;
    if ( list->count < list->bufSize ) list->pars [list->count++] = Parameter_vs;
    if ( list->count < list->bufSize ) list->pars [list->count++] = Parameter_dSquare;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_GirdleRatio;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_LowerFacets;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_Culet;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_TableDiameter;
    if ( list->count < list->bufSize ) list->pars [list->count++] = Parameter_vp;
    if ( list->count < list->bufSize ) list->pars [list->count++] = Parameter_Lh;
    if ( list->count < list->bufSize ) list->pars [list->count++] = Parameter_vLh;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_CuletX;
    if ( list->count < list->bufSize ) list->pars [list->count++] = DC_CuletY;
}

// Getting one parameter
static int PearGetParameter ( double & value, int param, PearData * data )
{
    switch (param)
    {
        default: return 0;
        case DC_CrownAngle_Table:
        case DC_CrownAngle:   value = data -> beta; break;
        case DC_PavilionAngle: value = data -> alph; break;
        case DC_Diameter: value = data->D; break;
        case DC_GirdleThickness: value = data->h; break;
        case Parameter_vs: value = data->vs; break;
        case DC_GirdleRatio: value = data->rt; break;
        case DC_LowerFacets: value = data->pavilion; break;
        case DC_Culet: value = data->culet; break;
        case DC_TableDiameter: value = data -> crown; break;
        case Parameter_vp: value = data -> vp; break;
        case Parameter_Lh: value = data -> Lh; break;
        case Parameter_vLh: value = data -> vLh; break;
        case DC_CuletX: value = data -> x; break;
        case DC_CuletY: value = data -> y; break;
        case Parameter_dSquare: value = data -> dSquare; break;
        // Dependent parameters
        case DC_CrownHeight : value = tan ( data -> beta ) * (1-data->crown) / 2; break;
        case DC_PavilionDepth: value = tan (data -> alph) * (1-data->culet) / 2; break;
        case DC_TotalHeight : value = tan ( data -> beta ) * (1-data->crown) / 2 +
            tan (data -> alph) * (1-data->culet) / 2 + data->h;
                        break;
    }
    return 1;
}

// Setting one parameter
static int PearSetParameter ( double value, int param, PearData * data )
{
    double v;
    switch (param)
    {
        default: return 0;
        case DC_CrownAngle_Table: data->beta = value; break;
        case DC_PavilionAngle: data -> alph = value; break;
        case DC_Diameter: data->D = value; break;
        case DC_GirdleThickness: data->h = value; break;
        case Parameter_vs: data->vs = value; break;
        case DC_GirdleRatio: data->rt = value; break;
        case DC_LowerFacets: data->pavilion = value; break;
        case DC_Culet: data->culet = value; break;
        case DC_TableDiameter: data -> crown = value; break;
        case Parameter_vp: data -> vp = value; break;
        case Parameter_Lh: data -> Lh = value; break;
        case Parameter_vLh: data -> vLh = value; break;
        case DC_CuletX: data -> x = value; break;
        case DC_CuletY: data -> y = value; break;
        case Parameter_dSquare: data -> dSquare = value; break;
        // Dependent parameters
        case DC_CrownHeight : v = tan (data -> beta);
                if (v > 1e-6) data -> crown = 1 - 2 * value / v;
                break;
        case DC_CrownAngle:
                if(value > 1e-6)
                {
                    data->crown = 1 - tan (data->beta) / tan (value) * (1-data->crown);
                    data->beta = value;
                }
                break;
        case DC_PavilionDepth: data->alph = atan2 ( value, (1 - data->culet) / 2 ); break;
    }
    return 1;
}

// The function of recalculating coordinates of 4th point for facets that are located one above
// another. We are trying to change Z coordinate first and only if it fails we will change
// X, Y coordinates. It is used for lower pavilion facets. 
static int MoveToPlanes ( const PHINDEX * S0, const PHINDEX * S1, POINT3 * vertex )
{
    POINT3 & p01 = vertex [S0 [0]];
    POINT3 & p02 = vertex [S0 [2]];
    // plane of the first facet
    VECTOR3 v0 = ( vertex [S0 [1]] - p01 ) & ( vertex [S0 [3]] - p01 );
    double d0 = -(p01|v0);

    POINT3 & p11 = vertex [S1 [0]];
    POINT3 & p12 = vertex [S1 [2]];
    // plane of the second facet
    VECTOR3 v1 = ( vertex [S1 [1]] - p11 ) & ( vertex [S1 [3]] - p11 );
    double d1 = -(p11|v1);

    // coordinates of p02 and p12 points coincide
    double a = (v0.x*p02.x + v0.y*p02.y) * v1.z -
             (v1.x*p02.x + v1.y*p02.y) * v0.z;
    double b = d1 * v0.z - d0 * v1.z;
    int answer = 1;
    double alpha = 1;
    double m0 = 1e-6;
    if ( fabs (a) < m0 )
    {
        if ( b >= 0 ) answer = 0;
    }
    else
    {
        double c = b / a;
        if ( a > 0 )
        {
            if ( c > 1 - m0 ) answer = 0;
            else if ( b > 0 ) alpha = c + m0;
        }
        else
        {
            if ( c < m0 ) answer = 0;
            else if ( b > a ) alpha = c - m0;
        }
    }
    if ( answer )
    {
        p02 *= alpha;
        p12 *= alpha;
        if ( fabs ( v0.z ) > 1e-8 ) p02.z = ( - d0 - (p02.x * v0.x + p02.y * v0.y) ) / v0.z;
        if ( fabs ( v1.z ) > 1e-8 ) p12.z = ( - d1 - (p12.x * v1.x + p12.y * v1.y) ) / v1.z;
    }
    else
    {
        answer = 0;
    }
    return answer;
}
